﻿
const settingsKey = sTID("settings");
const lastSettingsID = "easycut-lastSettings"
function saveLastSettings(settings) {
  saveString(lastSettingsID, settingsKey, JSON.stringify(settings))
}
function getLastSettings() {
  return getSavedString(lastSettingsID, settingsKey)
}
//------------------------------------------------
var menuSettings = {
  runInBackground: false,
  selectAfterwards: false,
  rasterizeSO: false,
  addNaming: true,
  mergePathComponents: true,
}

function getIds(strIds) {
  var ids = []
  for (var i = 0; i < strIds.length; i++) {
    ids.push(stringIDToTypeID(strIds[i]))
  }
  return JSON.stringify(ids)
}

function cutAtExistingPath() {
  onNewPathPoint(undefined, false)
}

function cutAtExistingSelection() {
  onNewSelection()
}

function cutLayerAtCoord(id, bounds, coord, horizontal, layerKind, namingCount) {
  var cutArea, cutArea2
  var b = bounds
  cutArea = horizontal ? [b[0], b[1], b[2], coord]
    : [b[0], b[1], coord, b[3]]
  if (layerKind != 1 && layerKind != 4) {
    cutArea2 = horizontal ? [b[0], coord, b[2], b[3]]
      : [coord, b[1], b[2], b[3]]
  }
  return applyCutting(id, layerKind, cutArea, cutArea2, namingCount)
}

function cutIntoParts(settings) {
  var hParts = settings.hCuts
  var vParts = settings.vCuts
  var layerIds = getSelectedLayersIds()

  app.activeDocument.suspendHistory('Cut into parts', '_cutParts()')

  function _cutParts() {
    var newIds = []
    for (var i = 0; i < layerIds.length; i++) {
      var id = layerIds[i]
      var desc
      try {
        var desc = layerDescById(id)
      } catch (e) { continue; }
      var b = getBoundsById(id)
      var layerKind = desc.getInteger(idlayerKind)
      var width = b[2] - b[0]
      var height = b[3] - b[1]
      var hStep = height / hParts
      var vStep = width / vParts

      var partToCut = id
      var partsIds = [id]
      for (var iH = 1; iH < hParts; iH++) { // cut horizontally
        var coord = Math.round(b[1] + hStep * iH)
        partsIds.push(cutLayerAtCoord(partToCut, b, coord, true, layerKind, iH))
      }

      var partsIds2 = []
      for (var p = 0; p < partsIds.length; p++) { // for each new part..
        partToCut = partsIds[p]
        for (var iV = 1; iV < vParts; iV++) { // cut vertically
          var coord = Math.round(b[0] + vStep * iV)
          partsIds2.push(cutLayerAtCoord(partToCut, b, coord, false, layerKind, iV))
        }
      }
      newIds = newIds.concat(partsIds.concat(partsIds2))
    }
    selectLayersByIdsCheck(newIds)
  }
}

function selectLayersByIdsCheck(ids) {
  if (menuSettings.selectAfterwards) {
    selectLayersByIds(ids)
  }
}

function cutAtExistingGuides() {
  var doc = app.activeDocument
  var guides = doc.guides

  app.activeDocument.suspendHistory('Cut at guides', '_cutExistingGuides()')

  function _cutExistingGuides() {
    var selectedIds = getSelectedLayersIds()
    for (var i = 0; i < guides.length; i++) {
      var guide = guides[i]
      var coord = Math.round(guide.coordinate)
      var horizontal = guide.direction == Direction.HORIZONTAL
      var newIds = cutAtAxis(coord, horizontal, selectedIds)
      selectedIds = selectedIds.concat(newIds)
    }
  }
}

function onNewSelection() {
  usePixelUnits()
  app.activeDocument.suspendHistory('Cut at selection', '_cutSelection()')
  function _cutSelection() {
    var selectedIds = getSelectedLayersIds()
    var newIds = cutAtSelection(selectedIds)
    selectLayersByIdsCheck(selectedIds.concat(newIds))
  }
  resetUnits()
}

function cutAtSelection(layerIds) {
  var selectionBounds = app.activeDocument.selection.bounds
  var channel = selectionToChannel("selection_to_cut")

  var newIds = []
  for (var i = 0; i < layerIds.length; i++) {
    var id = layerIds[i]
    var desc
    try {
      var desc = layerDescById(id)
    } catch (e) { continue; }
    var b = getBoundsById(id)

    if (!intersectBounds(b, selectionBounds)) {
      continue;
    }
    var layerKind = desc.getInteger(idlayerKind)
    newIds.push(applyCutting(id, layerKind, channel))
  }
  channel.remove()
  return newIds
}

function onSetShapePath() {
  var activePath = getActivePath()
  var pathKind = activePath.kind
  var subPathItem = activePath.subPathItems[activePath.subPathItems.length - 1]
  processPath(subPathItem, pathKind)
}

function onNewPathPoint(cutOn, fromEvent) {
  usePixelUnits()
  var fromEvent = fromEvent == undefined ? true : fromEvent
  var activePath = getActivePath()
  var pathKind = activePath.kind
  var onVectorMask = pathKind == PathKind.VECTORMASK
  var noNewPath = onVectorMask && activePath.subPathItems.length == 1
  if (!activePath || noNewPath) {
    return
  }
  var subPathItem = activePath.subPathItems[activePath.subPathItems.length - 1]
  var pathPoints = subPathItem.pathPoints

  if ((!cutOn || cutOn.onePointPath) && pathPoints.length === 1) {
    var anchor = pathPoints[0].anchor
    var left = pathPoints[0].leftDirection
    var right = pathPoints[0].rightDirection
    if (!pointsEqual(anchor, left) && !pointsEqual(anchor, right)) {
      processTwoPoints(left, right, 1, fromEvent, pathKind)
    }
  } else if ((!cutOn || cutOn.twoPointsPath) && pathPoints.length === 2) {
    if (pathPointIsUndirected(pathPoints[0]) && pathPointIsUndirected(pathPoints[1])) {
      processTwoPoints(pathPoints[0].anchor, pathPoints[1].anchor, 2, fromEvent, pathKind)
    }
  } else if ((!cutOn || cutOn.closedPath) && subPathItem.closed) {
    processPath(subPathItem, pathKind)
  }
  resetUnits()
}

function deleteActivePath(pathKind) {
  if (pathKind == PathKind.WORKPATH) {
    var desc = new ActionDescriptor();
    var ref = new ActionReference();
    var idWrPt = cTID("WrPt");
    ref.putProperty(idPath, idWrPt);
    desc.putReference(idnull, ref);
    executeAction(idDlt, desc, DialogModes.NO);
  } else {
    if (pathKind == PathKind.VECTORMASK) {
      var desc = new ActionDescriptor();
      var ref = new ActionReference();
      var idvectorMask = sTID("vectorMask");
      ref.putEnumerated(idPath, idPath, idvectorMask);
      desc.putReference(idnull, ref);
      executeAction(idslct, desc, DialogModes.NO);
    }
    executeAction(idDlt, undefined, DialogModes.NO);
  }
}

function processPath(subPath, pathKind) {
  var subPathInfo = subPathToInfo(subPath)
  var selectedIds = getSelectedLayersIds()
  deleteActivePath(pathKind)

  app.activeDocument.suspendHistory('Cut From Shape', '_cutShape()')
  function _cutShape() {
    var newIds = cutAtPath(subPathInfo, selectedIds)
    var idsToSelect = selectedIds.concat(newIds)
    selectLayersByIdsCheck(idsToSelect)
  }
}

function processTwoPoints(p1, p2, historyStatesToDelete, fromEvent, pathKind) {
  var selectedIds = getSelectedLayersIds()
  if (fromEvent) {
    for (var i = 0; i < historyStatesToDelete; i++) {
      deleteCurrentHistoryState()
    }
  } else {
    deleteActivePath(pathKind)
  }

  app.activeDocument.suspendHistory('Cut From Line', '_cut()')
  function _cut() {
    var newIds = cutAtLine(p1, p2, selectedIds)
    var idsToSelect = selectedIds.concat(newIds)
    selectLayersByIdsCheck(idsToSelect)
  }
}

function pathPointIsUndirected(p) {
  return pointsEqual(p.anchor, p.leftDirection) && pointsEqual(p.anchor, p.rightDirection)
}

function cutAtPath(subPathInfo, layerIds) {
  var newIds = []
  for (var i = 0; i < layerIds.length; i++) {
    var id = layerIds[i]
    var desc
    try {
      var desc = layerDescById(id)
    } catch (e) { continue; }
    var layerKind = desc.getInteger(idlayerKind)
    newIds.push(applyCutting(id, layerKind, subPathInfo))
  }
  return newIds
}


function cutAtLine(p, q, layerIds) {
  var newIds = []
  for (var i = 0; i < layerIds.length; i++) {
    var id = layerIds[i]
    var desc
    try {
      var desc = layerDescById(id)
    } catch (e) { continue; }

    var b = getBoundsById(id)
    var cut = lineBoundIntersection(b[0], b[1], b[2], b[3], p[0], p[1], q[0], q[1])
    if (!cut) {
      continue;
    }
    var layerKind = desc.getInteger(idlayerKind)
    var cutArea = getCutArea(cut, b)
    var cutArea2 = (layerKind != 1 && layerKind != 4) ? getCutAreaOpposite(cut, b) : []
    newIds.push(applyCutting(id, layerKind, cutArea, cutArea2))
  }
  return newIds
}

function getCutAreaOpposite(cut, b) {
  var cutArea = []
  var topLeft = [b[0], b[1]], topRight = [b[2], b[1]]
  var bottomRight = [b[2], b[3]], bottomLeft = [b[0], b[3]]
  if (cut.left && cut.right) {
    cutArea.push(cut.left, cut.right, bottomRight, bottomLeft)
  } else if (cut.left && cut.top) {
    cutArea.push(cut.top, topRight, bottomRight, bottomLeft, cut.left)
  } else if (cut.left && cut.bottom) {
    cutArea.push(cut.left, topLeft, topRight, bottomRight, cut.bottom)
  } else if (cut.right && cut.top) {
    cutArea.push(topLeft, cut.top, cut.right, bottomRight, bottomLeft)
  } else if (cut.right && cut.bottom) {
    cutArea.push(topLeft, topRight, cut.right, cut.bottom, bottomLeft)
  } else if (cut.top && cut.bottom) {
    cutArea.push(cut.top, topRight, bottomRight, cut.bottom)
  }
  return cutArea
}

function getCutArea(cut, b) {
  var cutArea = []
  var topLeft = [b[0], b[1]], topRight = [b[2], b[1]]
  var bottomRight = [b[2], b[3]], bottomLeft = [b[0], b[3]]
  if (cut.left && cut.right) {
    cutArea.push(topLeft, topRight, cut.right, cut.left)
  } else if (cut.left && cut.top) {
    cutArea.push(topLeft, cut.top, cut.left)
  } else if (cut.left && cut.bottom) {
    cutArea.push(cut.left, cut.bottom, bottomLeft)
  } else if (cut.right && cut.top) {
    cutArea.push(cut.top, topRight, cut.right)
  } else if (cut.right && cut.bottom) {
    cutArea.push(cut.right, bottomRight, cut.bottom)
  } else if (cut.top && cut.bottom) {
    cutArea.push(topLeft, cut.top, cut.bottom, bottomLeft)
  }
  return cutArea
}

function applyCutting(id, layerKind, cutArea, cutArea2, namingCount) {
  namingCount = namingCount || 1
  try {
    makeActiveById(id)
    switch (layerKind) {
      case 1: return cutFromNormalLayer(cutArea, namingCount)
      case 4: return cutFromShapeLayer(cutArea, namingCount)
      default: if(layerKind == 5 && menuSettings.rasterizeSO) {
        rasterizeActiveLayer()
        return cutFromNormalLayer(cutArea, namingCount)
      } else {
        return cutFromOtherLayer(cutArea, cutArea2, namingCount)
      }
    }
  } catch (e) { }
}

function onNewGuide() {
  usePixelUnits()
  var doc = app.activeDocument
  var guides = doc.guides
  var lastGuide = guides[guides.length - 1]
  var coord = Math.round(lastGuide.coordinate)
  var horizontal = lastGuide.direction == Direction.HORIZONTAL
  var selectedIds = getSelectedLayersIds()
  deleteCurrentHistoryState()
  app.activeDocument.suspendHistory('Cut Layer', '_cut()')

  function _cut() {
    var newIds = cutAtAxis(coord, horizontal, selectedIds)
    var idsToSelect = selectedIds.concat(newIds)
    selectLayersByIdsCheck(idsToSelect)
  }
  resetUnits()
}

function axisIntersectsBounds(bounds, coord, horizontal) {
  return (horizontal && bounds[1] < coord && bounds[3] > coord) ||
    (!horizontal && bounds[0] < coord && bounds[2] > coord)
}

function cutAtAxis(coord, horizontal, layerIds) {
  usePixelUnits()
  var newIds = []
  for (var i = 0; i < layerIds.length; i++) {
    var id = layerIds[i]
    var desc
    try {
      var desc = layerDescById(id)
    } catch (e) { continue; }
    var b = getBoundsById(id)
    var cutArea, cutArea2
    cutArea = horizontal ? [b[0], b[1], b[2], coord]
      : [b[0], b[1], coord, b[3]]

    if (!axisIntersectsBounds(b, coord, horizontal)) {
      continue;
    }

    var layerKind = desc.getInteger(idlayerKind)
    if (layerKind != 1 && layerKind != 4) {
      cutArea2 = horizontal ? [b[0], coord, b[2], b[3]]
        : [coord, b[1], b[2], b[3]]
    }

    newIds.push(applyCutting(id, layerKind, cutArea, cutArea2))
  }
  resetUnits()
  return newIds
}

/**
 * @param {*} area : area of type:
 * - Channel (of selection)
 * - SubPathInfo 
 * - bounds array (x1,y1,x2,y2)
 * - array of points ((x,y), (x,y,), ...)
 */
function selectArea(area) {
  var isChannel = false
  try { isChannel = area instanceof Channel } catch (e) { }
  if (area instanceof SubPathInfo) {
    var pathItem = createPathFromInfo(area)
    pathItem.makeSelection()
    pathItem.remove()
  } else if (isChannel) {
    activeDocument.selection.load(area, SelectionType.REPLACE)
  } else if (area[0].length) {
    var pathItem = createPath(area)
    pathItem.makeSelection()
    pathItem.remove()
  } else {
    select(area)
  }
}

function addCount(str, c) {
  return str + ' [' + c + ']'
}

function replaceCount(str, c) {
  return str.replace(/\[(.+?)\]/g, "[" + c + "]")
}

function cutFromNormalLayer(cutArea, namingCount) {
  var name = getActiveLayerName()
  setName(name, namingCount, 1)
  var hasStyles = copyLayerStyle()
  selectArea(cutArea)
  cutPixels()
  pastePixels(true)
  if(hasStyles) {
    pasteLayerStyle()
  }
  setName(name, namingCount)
  return getActiveLayerID()
}

function setName(str, num, add) {
  if (menuSettings.addNaming) {
    var n = add ? num + add : num
    str = num > 1 ? replaceCount(str, n) : addCount(str, n)
  }
  setActiveLayerName(str)
}

function cutFromOtherLayer(cutArea, cutArea2, namingCount) {
  var name = getActiveLayerName()
  setName(name, namingCount)
  duplicateActive()
  selectArea(cutArea)
  createNewMaskFromSelection()
  selectLowerLayer()
  setName(name, namingCount, 1)
  if (cutArea2) {
    selectArea(cutArea2)
  } else {
    selectArea(cutArea)
    invertSelection()
  }
  createNewMaskFromSelection()
  return getActiveLayerID()
}

function pointArrayToSubPathInfo(ptArr, shapeOperation) {
  var subPathInfo = new SubPathInfo();
  shapeOperation = shapeOperation || ShapeOperation.SHAPEADD
  var points = []
  for (var i = 0; i < ptArr.length; i++) {
    var p = ptArr[i]
    points[i] = new PathPointInfo;
    points[i].kind = PointKind.CORNERPOINT;
    points[i].anchor = p;
    points[i].leftDirection = p;
    points[i].rightDirection = p;
  }
  subPathInfo.closed = true;
  subPathInfo.operation = shapeOperation
  subPathInfo.entireSubPath = points;
  return subPathInfo
}

function selectVectorMask() {
  var desc = new ActionDescriptor();
  var ref = new ActionReference();
  var idvectorMask = sTID("vectorMask");
  ref.putEnumerated(idPath, idPath, idvectorMask);
  ref.putEnumerated(idLyr, idOrdn, idTrgt);
  desc.putReference(idnull, ref);
  executeAction(idslct, desc, DialogModes.NO);
}

function getActivePathInfo() {
  var activePath = getActivePath()
  var subPaths = []
  for (var i = 0; i < activePath.subPathItems.length; i++) {
    subPaths.push(subPathToInfo(activePath.subPathItems[i]))
  }
  return subPaths
}

function getAreaSubPathInfo(area) {
  var isChannel = false
  try { isChannel = area instanceof Channel } catch (e) { }
  if (area instanceof SubPathInfo) {
    return area
  } else if (isChannel) {
    activeDocument.selection.load(area, SelectionType.REPLACE)
    pathFromSelection()
    return getActivePathInfo()[0]
  } else if (area[0].length) {
    return pointArrayToSubPathInfo(area)
  } else {
    return boundsToSubPathInfo(area)
  }
}

function pointInfo(p) {
  var i = new PathPointInfo
  i.kind = PointKind.CORNERPOINT
  i.anchor = p
  i.leftDirection = p
  i.rightDirection = p
  return i
}

function boundsToSubPathInfo(b) {
  var subPathInfo = new SubPathInfo();
  var points = [
    pointInfo([b[0], b[1]]),
    pointInfo([b[2], b[1]]),
    pointInfo([b[2], b[3]]),
    pointInfo([b[0], b[3]]),
  ]
  subPathInfo.closed = true;
  subPathInfo.entireSubPath = points;
  return subPathInfo
}

function selectPenTool() {
  var desc = new ActionDescriptor();
  var ref = new ActionReference();
  var idpenTool = sTID("penTool");
  ref.putClass(idpenTool);
  desc.putReference(idnull, ref);
  var iddontRecord = sTID("dontRecord");
  desc.putBoolean(iddontRecord, true);
  var idforceNotify = sTID("forceNotify");
  desc.putBoolean(idforceNotify, true);
  executeAction(idslct, desc, DialogModes.NO);
}

function addPathInfoToVectorMask(pathInfo, operation) {
  pathInfo.operation = operation
  var cutPath = activeDocument.pathItems.add('shapePart', [pathInfo]);
  executeAction(idcopy, undefined, DialogModes.NO);
  cutPath.remove()
  selectVectorMask()
  executeAction(idpast, undefined, DialogModes.NO);
  if (menuSettings.mergePathComponents) {
    combineShapeParts()
  }
}

function cutFromShapeLayer(cutArea, namingCount) {
  var activeTool = getToolID()
  var name = getActiveLayerName()
  selectPenTool()
  var cutInfo = getAreaSubPathInfo(cutArea)
  duplicateActive2()
  addPathInfoToVectorMask(cutInfo, ShapeOperation.SHAPEINTERSECT)
  var newId = getActiveLayerID()
  setName(name, namingCount)
  selectLowerLayer()
  addPathInfoToVectorMask(cutInfo, ShapeOperation.SHAPESUBTRACT) 
  setName(name, namingCount, 1)
  setTool(activeTool)
  return newId
}

function lineBoundIntersection(xmin, ymin, xmax, ymax, x1, y1, x2, y2) {
  var intersection = {}
  // intersections with Bounds
  var y_left = y1 + (y2 - y1) * (xmin - x1) / (x2 - x1)
  var y_right = y1 + (y2 - y1) * (xmax - x1) / (x2 - x1)
  var x_top = x1 + (x2 - x1) * (ymin - y1) / (y2 - y1)
  var x_bottom = x1 + (x2 - x1) * (ymax - y1) / (y2 - y1)

  if (!isFinite(y_left) || !isFinite(x_top)) {
    return undefined // no intersection
  }

  if (ymin <= y_left && y_left <= ymax) {
    intersection.left = [xmin, y_left]
  }
  if (ymin <= y_right && y_right <= ymax) {
    intersection.right = [xmax, y_right]
  }
  if (xmin <= x_top && x_top <= xmax) {
    intersection.top = [x_top, ymin]
  }
  if (xmin <= x_bottom && x_bottom <= xmax) {
    intersection.bottom = [x_bottom, ymax]
  }
  return intersection
}

function pointsEqual(p1, p2) {
  return p1[0] == p2[0] && p1[1] == p2[1]
}





